# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .accesscode import AccessCode
from .domainname import DomainName
from .emailaddress import EmailAddress
from .genericdeliverypointspecification import GenericDeliveryPointSpecification
from .genericpostaladdress import GenericPostalAddress
from .iso3166 import ISO3166Alpha2
from .list_ import List
from .objectmeta import ObjectMeta
from .phonenumber import Phonenumber
from .romanizedname import RomanizedName
from .resource import Resource
from .resourceevent import ResourceEvent
from .resourceeventtype import ResourceEventType
from .resourcemetaclass import Config as ResourceConfig
from .signature import Signature
from .stringtype import StringType
from .symbolicname import SymbolicName


__all__: list[str] = [
    'AccessCode',
    'DomainName',
    'EmailAddress',
    'GenericDeliveryPointSpecification',
    'GenericPostalAddress',
    'ISO3166Alpha2',
    'List',
    'ObjectMeta',
    'Phonenumber',
    'Resource',
    'ResourceConfig',
    'ResourceEvent',
    'ResourceEventType',
    'RomanizedName',
    'Signature',
    'StringType',
    'SymbolicName',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import date
from datetime import datetime

import pydantic

from .shipmenttrackingnumber import ShipmentTrackingNumber


class ShipmentTrackingInfo(pydantic.BaseModel):
    """Contains information to track the status of a :term:`Shipment`."""
    carrier_id: str = pydantic.Field(
        default=...,
        title="Carrier ID",
        description="An identifier for the the Carrier, usually a domain name."
    )

    carrier_name: str = pydantic.Field(
        default=...,
        title="Carrier",
        description=(
            "The human-readable name of the carrier that is providing the "
            "shipping services."
        )
    )

    tracking_numbers: list[ShipmentTrackingNumber] = pydantic.Field(
        default=[],
        title="Track & Trace",
        description=(
            "An array of tracking numbers related to the shipments. Some "
            "carriers issue a tracking number for each package in a "
            "shipment."
        )
    )

    etd: date | datetime = pydantic.Field(
        default=...,
        title="Estimated Time of Departure (ETD)",
        description=(
            "Expected Time of Departure is the prediction of time "
            "that is expected for a transport system to depart "
            "its point of origin or location. As shipping labels "
            "are usually created prior to actual shipment, this "
            "estimation should be interpreted as *on or before*."
        )
    )

    @property
    def tracking_number(self) -> str | None:
        """The tracking number if the :term:`Shipment` has a single
        tracking number, else ``None``. If :attr:`tracking_number`
        is ``None``, inspect the :attr:`tracking_numbers` attribute
        for the tracking codes.
        """
        if len(self.tracking_numbers) > 1\
        or not self.tracking_numbers:
            return None
        return self.tracking_numbers[0].number
    
    def add(self, trackingcode: str, trackingurl: str | None) -> None:
        if self.has(trackingcode):
            return
        self.tracking_numbers.append(
            ShipmentTrackingNumber(
                number=trackingcode,
                url=trackingurl
            )
        )

    def has(self, trackingcode: str) -> bool:
        """Return a boolean indicating if the tracking code is
        included in this shipment.
        """
        return any([x.number for x in self.tracking_numbers])


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Literal

import pydantic

from .listmeta import ListMeta


class _EntityList:

    def __init__(self, getitem: Callable[..., Any]):
        self._getitem = getitem

    def __getitem__(self, model: type[pydantic.BaseModel]) -> pydantic.BaseModel:
        return self._getitem(model)


@_EntityList
def List(model: type[pydantic.BaseModel]):
    name = f'{model.__name__}List'
    return type(name, (pydantic.BaseModel,), {
        'api_version': pydantic.Field(
            default='v1',
            alias='apiVersion',
            title='API Version',
            description=(
                "APIVersion defines the versioned schema of this representation of "
                "an object. Servers should convert recognized schemas to the latest "
                "internal value, and may reject unrecognized values."
            )
        ),
        'kind': pydantic.Field(
            default=name,
            title='Kind',
            description=(
                "Kind is a string value representing the REST resource this "
                "object represents. Servers may infer this from the endpoint "
                "the client submits requests to. Cannot be updated. In "
                "CamelCase. "
            )
        ),
        'metadata': pydantic.Field(
            default_factory=ListMeta,
            title='Metadata',
            description="Standard list metadata."
        ),
        'items': pydantic.Field(
            default=[],
            title='Items',
            description=f'List of **{model.__name__}** objects.'
        ),
        '__annotations__': {
            'api_version': str,
            'kind': Literal[f'{model.__name__}List'],
            'metadata': ListMeta,
            'items': list[model]
        }
    })


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import warnings
from typing import Any
from typing import Callable
from typing import Generator

try:
    import iso3166
except ImportError:
    warnings.warn(
        "The iso3166 module is not installed. No validation is "
        "performed on types that assume any ISO 3166 code."
    )
    iso3166 = None
from pydantic.validators import str_validator


# Corrections for wrong names
COUNTRY_NAMES: dict[str, str] = {
    'NL': 'The Netherlands'
}


class ISO3166Base(str):
    length: int

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[..., str | None], None, None]:
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, v: str) -> str:
        if len(v) != cls.length:
            raise ValueError('Value is not an ISO 3166 code.')
        try:
            return cls(cls._get_country(v))
        except LookupError:
            raise ValueError(f"Invalid code: {v}")

    @classmethod
    def _get_country(cls, v: str) -> str:
        raise NotImplementedError


class ISO3166Alpha2(ISO3166Base):
    length: int = 2

    @property
    def name(self) -> str:
        if str(self) in COUNTRY_NAMES:
            return COUNTRY_NAMES[self]
        c = iso3166.countries_by_alpha2.get(self)
        return c.name

    @classmethod
    def __modify_schema__(
        cls,
        field_schema: dict[str, Any]
    ) -> None:
        field_schema.update( # pragma: no cover
            title='Country code',
            description='An ISO 3166 alpha-2 country code.',
            type='string',
        )

    @classmethod
    def _get_country(cls, v: str) -> str:
        if iso3166 is None:
            return v
        c = iso3166.countries_by_alpha2.get(v)
        if c is None:
            raise LookupError
        return c.alpha2


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


class ShipmentTrackingNumber(pydantic.BaseModel):
    number: str = pydantic.Field(
        default=...,
        title="Number",
        description=(
            "An identifier issued by a Carrier to reference "
            "a Shipment."
        )
    )

    url: str | None = pydantic.Field(
        default=None,
        title="URL",
        description=(
            "A URL pointing to a web page where more information regarding the "
            "Shipment may be retrieved."
        )
    )

    def __str__(self) -> str:
        return self.number


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


class Signature(pydantic.BaseModel):
    """A digital signature establishing the non-repudiation of a
    transaction.
    """
    alg: str = pydantic.Field(
        default=...,
        title="Algorithm",
        description=(
            "The signing algorithm used to create `signature`. Must be "
            "a valid JSON Web Algorithm (JWA)."
        )
    )

    kid: str = pydantic.Field(
        default=...,
        title="Key identifier",
        description=(
            "The key identifier of the signing key. A key identifier is "
            "generated as follows:\n\n"
            "- For keys using the **RSA** algorithm, the key identifier is "
            "the URL Safe Base64-encoded SHA-256 hash of the concatenation "
            "of the ASCII-encoded string `RSA`, `e`, and `n`. The value of "
            "`e` is encoded as a 3 octets and `n` is encoded in the number "
            "of octets of the public modulus. The expected byte order is "
            "big-endian (BE).\n\n"
            "- For keys using the **Edwards curves**, the key identifier is "
            "the URL Safe Base64-encoded SHA-256 hash of the concatenation "
            "of the ASCII-encoded string `OKP`, then the ASCII-encoded string "
            "containing the curve name (either `Ed448` or `Ed25519`, "
            "case-sensitive), and then a byte-sequence containing the public "
            "key in 'raw' encoding *and* format. The expected byte order is "
            "big-endian (BE)."
        )
    )

    digest: str = pydantic.Field(
        default=...,
        title="Digest",
        description=(
            "The URL Safe Base64-encoded SHA-384 hash of the transaction. "
            "Refer to the documentation of the specific transaction to "
            "determine how to create the digest."
        )
    )

    signature: str = pydantic.Field(
        default=...,
        title="Signature",
        description=(
            "The digital signature created with the keypair identified by "
            "`kid` from `digest`."
        )
    )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
import pytest

from canonical import EmailAddress


@pytest.mark.parametrize('v', [
    'abc@example.com',
])
def test_validate(v: str):
    EmailAddress.validate(v)


def test_as_model_attribute():
    class Model(pydantic.BaseModel):
        email: EmailAddress

    Model.parse_obj({'email': 'abc@example.com'})


def test_domain():
    e = EmailAddress.validate('abc@example.com')
    assert e.domain == 'example.com'



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Any
from typing import TypeVar

import pydantic


T = TypeVar('T', bound=pydantic.BaseModel)


class ResourceEvent(pydantic.BaseModel):
    kind: str

    message: str | None = pydantic.Field(
        default=None,
        title='Message',
        description='An optional message describing the event.'
    )

    timestamp: datetime.datetime = pydantic.Field(
        default_factory=datetime.datetime.utcnow
    )

    @classmethod
    def observe(
        cls: type[T],
        **params: Any
    ) -> T:
        return cls.parse_obj(params)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from types import EllipsisType
from typing import Any
from typing import Callable
from typing import Generator
from typing import Pattern
from typing import TypeVar

from pydantic.validators import str_validator


T = TypeVar('T')


class SymbolicNameMetaclass(type):

    def __getitem__(
        self: T,
        opts: int | tuple[
            EllipsisType | int | None,
            EllipsisType | int | None,
            EllipsisType | str | None,
        ],
    ) -> T:
        min_length: int | EllipsisType | None = 6
        max_length: int | EllipsisType | None = 31
        pattern: None | Pattern[str] = re.compile('^[a-z][a-z0-9\\-]+[a-z0-9]$')
        if isinstance(opts, int):
            max_length = opts
        elif len(opts) >= 2:
            min_length, max_length, *remaining = opts
            if remaining:
                if isinstance(remaining[0], str):
                    pattern = re.compile(remaining[0])
                elif remaining[0] is None:
                    pattern = None
        if min_length == ...:
            min_length = 6
        if min_length is None:
            min_length = 0
        if max_length == ...:
            max_length = 31
        if max_length is None:
            max_length = 0
        assert min_length < max_length # nosec
        return type(self.__name__, (self,), { # type: ignore
            'min_length': min_length,
            'max_length': max_length,
            'pattern': pattern
        })


class SymbolicName(str, metaclass=SymbolicNameMetaclass):
    min_length: int = 6
    max_length: int = 31
    pattern: Pattern[str] | None = re.compile('^[a-z][a-z0-9\\-]+[a-z0-9]$')

    @classmethod
    def __modify_schema__(
        cls,
        field_schema: dict[str, Any]
    ) -> None:
        field_schema.update( # pragma: no cover
            title='Symbolic Name',
            type='string',
        )

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[..., str | None], None, None]:
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, v: str) -> str:
        if cls.min_length and len(v) < cls.min_length:
            raise ValueError("Symbolic name is too short.")
        if cls.max_length and len(v) > cls.max_length:
            raise ValueError("Symbolic name is too long.")
        if cls.pattern and not cls.pattern.match(v):
            raise ValueError("Invalid characters.")
        return cls(v)

    def __repr__(self) -> str: # pragma: no cover
        return f'SymbolicName({self})'